@@ -17,7 +17,7 @@ from tethys_utils.data_models import Geometry, Dataset, DatasetBase, Station, St
 from tethys_utils.misc import make_run_date_key, grp_ts_agg, write_pkl_zstd
 # from misc import make_run_date_key, grp_ts_agg, write_pkl_zstd
 import geojson
-from shapely.geometry import shape, mapping
+from shapely.geometry import shape, mapping, Point, box
 from shapely import wkb
 
 ############################################
@@ -31,47 +31,59 @@ agg_stat_mapping = {'mean': 'mean', 'cumulative': 'sum', 'continuous': None, 'ma
 ### Functions
 
 
-def results_data_integrety_checks(data, param_name, attrs, encoding, spatial_distribution, ancillary_variables=None):
+def extract_data_dimensions(data, parameter):
     """
 
     """
-    # Time series data
-    data_cols = []
-    if isinstance(ancillary_variables, list):
-        if len(ancillary_variables) > 0:
-            for av in ancillary_variables:
-                if not av in data:
-                    raise ValueError('The DataFrame must contain every value in the ancillary_variables list')
-                else:
-                    data_cols.extend([av])
+    data_index = data[parameter].dims
+    vars2 = list(data.variables)
+    vars3 = [v for v in vars2 if v not in data_index]
 
-    data_cols.extend([param_name])
+    vars_dict = {}
+    for v in vars3:
+        index1 = data[v].dims
+        vars_dict[v] = index1
 
-    if spatial_distribution == 'sparse':
+    ancillary_variables = [v for v, i in vars_dict.items() if (i == data_index) and (v != parameter)]
+
+    main_vars = [parameter] + ancillary_variables
+
+    return data_index, vars3, main_vars, ancillary_variables, vars_dict
+
+
+def data_integrety_checks(data, parameter, ds_metadata, attrs, encoding):
+    """
+
+    """
+    # data dimensions
+    data_index, vars1, main_vars, ancillary_variables, vars_dict = extract_data_dimensions(data, parameter)
+
+    sd = ds_metadata['spatial_distribution']
+    geo_type = ds_metadata['geometry_type']
+    grouping = ds_metadata['grouping']
+
+    if ('geometry' in data_index) and (sd == 'grid'):
+        raise ValueError('You have passed geometry as an index, but the spatial_distribution is labeled as grid...something is inconsistent.')
+    elif (sd == 'sparse') and (('lon' in data_index) or ('lat' in data_index)):
+        raise ValueError('You have passed lon/lat as an index, but the spatial_distribution is labeled as sparse...something is inconsistent.')
+
+    if sd == 'sparse':
         ts_index_list = ['geometry', 'time', 'height']
-    elif spatial_distribution == 'grid':
+    elif sd == 'grid':
         ts_index_list = ['lon', 'lat', 'time', 'height']
     else:
         raise ValueError('spatial_distribution should be either sparse or grid.')
 
-    ts_essential_list = [param_name]
-    ts_no_attrs_list = ['modified_date', 'lat', 'lon', 'geometry']
-
-    ts_data_cols = list(data.columns)
-
-    for c in ts_essential_list:
-        if not c in ts_data_cols:
-            raise ValueError('The DataFrame must contain the column: ' + str(c))
-
-    ts_data_index = list(data.index.names)
+    ts_essential_list = [parameter]
+    ts_no_attrs_list = ['modified_date', 'station_id', 'lat', 'lon', 'name', 'altitude', 'ref', 'virtual_station', 'geometry']
 
     for c in ts_index_list:
-        if not c in ts_data_index:
-            raise ValueError('The DataFrame must contain the index: ' + str(c))
+        if not c in data_index:
+            raise ValueError('The Data must contain the dimension: ' + str(c))
 
     if isinstance(attrs, dict):
         attrs_keys = list(attrs.keys())
-        for col in ts_data_cols:
+        for col in vars1:
             if not col in ts_no_attrs_list:
                 if not col in ts_essential_list:
                     if not col in attrs_keys:
@@ -80,70 +92,15 @@ def results_data_integrety_checks(data, param_name, attrs, encoding, spatial_dis
         raise TypeError('attrs must be a dict')
 
     if isinstance(encoding, dict):
-        for col in ts_data_cols:
+        for col in vars1:
             if not col in ts_no_attrs_list:
                 if not col in encoding:
                     raise ValueError(col + ' must be in the encoding dict')
 
-    return data
-
-
-def station_data_integrety_checks(data, attrs=None, encoding=None):
-    """
-
-    """
-    stn_m = StationBase(**data)
-    stn_data = orjson.loads(stn_m.json(exclude_none=True))
-
-    essential_keys = ['geometry', 'station_id']
-
-    for c in essential_keys:
-        if not c in stn_data:
-            raise ValueError('The Dict must contain the key: ' + str(c))
-
-
-    ignore_attr_keys = ['station_id', 'lat', 'lon', 'name', 'altitude', 'ref', 'virtual_station', 'geometry']
-    other_keys = [s for s in stn_data if s not in ignore_attr_keys]
-
-    if other_keys:
-
-        if isinstance(attrs, dict):
-            for col in other_keys:
-                if not col in attrs:
-                    raise ValueError(col + ' must be in the attrs dict')
-        else:
-            raise TypeError('attrs must be a dict')
-
-        if isinstance(encoding, dict):
-            for col in other_keys:
-                if not col in encoding:
-                    raise ValueError(col + ' must be in the encoding dict')
-        else:
-            raise TypeError('encoding must be a dict')
-
-    return stn_data
-
-
-def metadata_integrety_checks(results_data, station_data, ds_metadata):
-    """
-
-    """
-    sd = ds_metadata['spatial_distribution']
-    geo_type = ds_metadata['geometry_type']
-    grouping = ds_metadata['grouping']
-
-    data_index = list(results_data.index.names)
-    if ('geometry' in data_index) and (sd == 'grid'):
-        raise ValueError('You have passed geometry as an index, but the spatial_distribution is labeled as grid...something is inconsistent.')
-    elif (sd == 'sparse') and (('lon' in data_index) or ('lat' in data_index)):
-        raise ValueError('You have passed lon/lat as an index, but the spatial_distribution is labeled as sparse...something is inconsistent.')
-
-    stn_geometry = station_data['geometry']
-    stn_geo_type = stn_geometry['type']
-
+    ## Geometry and metadata checks
     if 'geometry' in data_index:
         data_sd = 'sparse'
-        hex_geo1 = results_data.index.get_level_values('geometry').unique()
+        hex_geo1 = np.unique(data['geometry'])
         if len(hex_geo1) == 1:
             data_grouping = 'none'
         else:
@@ -154,8 +111,8 @@ def metadata_integrety_checks(results_data, station_data, ds_metadata):
     elif ('lon' in data_index) and ('lat' in data_index):
         data_sd = 'grid'
         data_geo_type = 'Point'
-        lons = results_data.index.get_level_values('lon').unique()
-        lats = results_data.index.get_level_values('lat').unique()
+        lons = np.unique(data['lon'])
+        lats = np.unique(data['lat'])
         if (len(lons) == 1) and (len(lats) == 1):
             data_grouping = 'none'
         else:
@@ -163,12 +120,17 @@ def metadata_integrety_checks(results_data, station_data, ds_metadata):
     else:
         raise ValueError('Data is not indexed correctly.')
 
-    if data_grouping == 'blocks':
-        if stn_geo_type != 'Polygon':
-            raise ValueError('If the data grouping is "blocks", then the station geometry should be a Polygon rectangle of the extent.')
-    else:
-        if stn_geo_type != data_geo_type:
-            raise ValueError('If the data grouping is "none", then the station geometry should be the same as the data geometry.')
+    # Check for station_id
+    if data_grouping != 'blocks':
+        if 'station_id' not in vars1:
+            raise ValueError('If the grouping is not blocks, then the station_id should be in the data.')
+
+    # if data_grouping == 'blocks':
+    #     if 'extent' not in data_index:
+    #         raise ValueError('If the data grouping is "blocks", then the Dataset should include a Polygon rectangle of the extent.')
+    # else:
+    #     if stn_geo_type != data_geo_type:
+    #         raise ValueError('If the data grouping is "none", then the station geometry should be the same as the data geometry.')
 
     if data_geo_type != geo_type:
         raise ValueError('The data geometry type does not match the geometry type listed in the dataset metadata.')
@@ -177,8 +139,10 @@ def metadata_integrety_checks(results_data, station_data, ds_metadata):
     if data_sd != sd:
         raise ValueError('The data spatial distribution does not match the spatial distribution listed in the dataset metadata.')
 
+    return data.copy()
 
-def data_to_xarray(results_data, station_data, param_name, results_attrs, results_encoding, station_attrs=None, station_encoding=None, run_date=None, ancillary_variables=None, compression=False, compress_level=1):
+
+def package_xarray(data, parameter, attrs, encoding, run_date=None, compression=False, compress_level=1):
     """
     Converts DataFrames of time series data, station data, and other attributes to an Xarray Dataset. Optionally has Zstandard compression.
 
@@ -204,130 +168,98 @@ def data_to_xarray(results_data, station_data, param_name, results_attrs, result
     Xarray Dataset or bytes object
     """
     ## dataset metadata
-    ds_mapping = results_attrs[param_name]
-    param_name = ds_mapping['parameter']
-    sd = ds_mapping['spatial_distribution']
-    grouping = ds_mapping['grouping']
-    geo_type = ds_mapping['geometry_type']
+    ds_metadata = attrs[parameter]
+    sd = ds_metadata['spatial_distribution']
+    grouping = ds_metadata['grouping']
+    # geo_type = ds_metadata['geometry_type']
 
     ## Integrity Checks
+    data1 = data_integrety_checks(data, parameter, ds_metadata, attrs, encoding)
 
-    # Time series data
-    ts_data1 = results_data_integrety_checks(results_data, param_name, results_attrs, results_encoding, sd, ancillary_variables)
+    ## data dimensions
+    data_index, vars1, main_vars, ancillary_variables, vars_dict = extract_data_dimensions(data, parameter)
 
-    # Station data
-    stn_data = station_data_integrety_checks(station_data, attrs=station_attrs, encoding=station_encoding)
-
-    # Metadata
-    metadata_integrety_checks(ts_data1, stn_data, ds_mapping)
-
-    ## geometry from station data
-    stn_geometry = stn_data['geometry']
-    stn_data.pop('geometry')
+    ## Create extent and staion_id if grouping is blocks
     if grouping == 'blocks':
-        stn_data['extent'] = shape(stn_geometry).wkb_hex
+        if sd == 'grid':
+            res = np.mean(np.diff(np.unique(data1['lon'])))
+            min_x = (data1['lon'].min() - (res*0.5)).round(7)
+            max_x = (data1['lon'].max() + (res*0.5)).round(7)
+            min_y = (data1['lat'].min() - (res*0.5)).round(7)
+            max_y = (data1['lat'].max() + (res*0.5)).round(7)
+        else:
+            geos1 = [wkb.loads(g, True) for g in data1['geometry'].values]
+            # TODO: Finish the lines and polygon extent creation
 
-    ## Station properties
-    if 'properties' in stn_data:
-        props = stn_data['properties'].copy()
-        for k, v in props.items():
-            stn_data[k] = v
-        stn_data.pop('properties')
+        extent = box(min_x, min_y, max_x, max_y)
+        extent_geo = mapping(extent)
+        stn_id = assign_station_id(extent_geo)
+        extent_hex = extent.wkb_hex
+        data1 = data1.assign_coords([extent_hex])
+        data1 = data1.assign({'station_id': (('extent'), [stn_id])})
 
     ## Assign Attributes
     attrs1 = {'station_id': {'cf_role': "timeseries_id"}, 'lat': {'standard_name': "latitude", 'units': "degrees_north"}, 'lon': {'standard_name': "longitude", 'units': "degrees_east"}, 'altitude': {'standard_name': 'surface_altitude', 'long_name': 'height above the geoid to the lower boundary of the atmosphere', 'units': 'm'}, 'geometry': {'long_name': 'The hexadecimal encoding of the Well-Known Binary (WKB) geometry', 'crs_EPSG': 4326}, 'extent': {'long_name': 'The hexadecimal encoding of the Well-Known Binary (WKB) station extent', 'crs_EPSG': 4326}}
 
-    if 'name' in stn_data.keys():
+    if 'name' in vars1:
         attrs1.update({'name': {'long_name': 'station name'}})
-    if 'ref' in stn_data.keys():
+    if 'ref' in vars1:
         attrs1.update({'ref': {'long_name': 'station reference id given by the owner'}})
 
-    if isinstance(station_attrs, dict):
-        attrs1.update(copy.deepcopy(station_attrs))
+    if isinstance(attrs, dict):
+        attrs1.update(copy.deepcopy(attrs))
 
-    ts_cols = list(ts_data1.columns)
-
-    attrs1.update(results_attrs)
-    if 'cf_standard_name' in attrs1[param_name]:
-        attrs1[param_name]['standard_name'] = attrs1[param_name].pop('cf_standard_name')
+    if 'cf_standard_name' in attrs1[parameter]:
+        attrs1[parameter]['standard_name'] = attrs1[parameter].pop('cf_standard_name')
     attrs1.update({'height': {'standard_name': 'height', 'long_name': 'vertical distance above the surface', 'units': 'm', 'positive': 'up'}, 'time': {'standard_name': 'time', 'long_name': 'start_time'}})
-    if 'modified_date' in ts_cols:
+    if 'modified_date' in vars1:
         attrs1.update({'modified_date': {'long_name': 'last modified date'}})
 
     if isinstance(ancillary_variables, list):
         if len(ancillary_variables) > 0:
-            attrs1[param_name].update({'ancillary_variables': ' '.join(ancillary_variables)})
+            attrs1[parameter].update({'ancillary_variables': ' '.join(ancillary_variables)})
 
     ## Assign encodings
     encoding1 = {'lon': {'dtype': 'int32', '_FillValue': -999999, 'scale_factor': 0.0000001}, 'lat': {'dtype': 'int32', '_FillValue': -999999, 'scale_factor': 0.0000001}, 'altitude': {'dtype': 'int32', '_FillValue': -9999, 'scale_factor': 0.001}}
 
-    if isinstance(station_encoding, dict):
-        encoding1.update(copy.deepcopy(station_encoding))
+    if isinstance(encoding, dict):
+        encoding1.update(copy.deepcopy(encoding))
 
-    height = pd.to_numeric(ts_data1.reset_index()['height'], downcast='integer')
+    data1['height'] = pd.to_numeric(data1['height'].values, downcast='integer')
 
-    if 'int' in height.dtype.name:
-        height_enc = {'dtype': height.dtype.name}
-    elif 'float' in height.dtype.name:
+    if 'int' in data1['height'].dtype.name:
+        height_enc = {'dtype': data1['height'].dtype.name}
+    elif 'float' in data1['height'].dtype.name:
         height_enc = {'dtype': 'int32', '_FillValue': -9999, 'scale_factor': 0.001}
     else:
         raise TypeError('height should be either an int or a float')
 
-    encoding1.update(results_encoding)
     encoding1.update({'time': {'_FillValue': -99999999, 'units': "days since 1970-01-01 00:00:00"}, 'height': height_enc})
-    if 'modified_date' in ts_cols:
+    if 'modified_date' in vars1:
         encoding1.update({'modified_date': {'_FillValue': -99999999, 'units': "days since 1970-01-01 00:00:00"}})
 
-    ## Create the Xarray Dataset
-    data_index = list(ts_data1.index.names)
-    ds1 = ts_data1.to_xarray()
-
-    ## Assign the stn data to the main Dataset
-    if 'extent' in stn_data:
-        ds1 = ds1.assign_coords({'extent': [stn_data['extent']]})
-        stn_data.pop('extent')
-
-    if 'extent' in ds1:
-        stn_index = ('extent')
-    else:
-        if 'geometry' in data_index:
-            stn_index = ('geometry')
-        else:
-            stn_index = ('lon', 'lat')
-
-    # ds1 = ds1.assign_coords({'station_id': (stn_index, [stn_data['station_id']])})
-    # stn_data.pop('station_id')
-
-    for k, v in stn_data.items():
-        ds1 = ds1.assign({k: (stn_index, [v])})
-
-    ## Assign the lat and lon if the data are sparse points
-    if ('geometry' in data_index) and (stn_geometry['type'] == 'Point'):
-        ds1 = ds1.assign({'lon': (stn_index, [stn_geometry['coordinates'][0]])})
-        ds1 = ds1.assign({'lat': (stn_index, [stn_geometry['coordinates'][1]])})
-
     ## Add attributes and encodings
     for e, val in encoding1.items():
-        if e in ds1:
+        if e in data1:
             if ('dtype' in val) and (not 'scale_factor' in val):
                 if 'int' in val['dtype']:
-                    ds1[e] = ds1[e].astype(val['dtype'])
+                    data1[e] = data1[e].astype(val['dtype'])
             if 'scale_factor' in val:
                 precision = int(np.abs(np.log10(val['scale_factor'])))
-                ds1[e] = ds1[e].round(precision)
-            ds1[e].encoding = val
+                data1[e] = data1[e].round(precision)
+            data1[e].encoding = val
 
     for a, val in attrs1.items():
-        if a in ds1:
-            ds1[a].attrs = val
+        if a in data1:
+            data1[a].attrs = val
 
-    title_str = '{agg_stat} {parameter} in {units} of the {feature} by a {method} owned by {owner}'.format(agg_stat=ds_mapping['aggregation_statistic'], parameter=ds_mapping['parameter'], units=ds_mapping['units'], feature=ds_mapping['feature'], method=ds_mapping['method'], owner=ds_mapping['owner'])
+    title_str = '{agg_stat} {parameter} in {units} of the {feature} by a {method} owned by {owner}'.format(agg_stat=ds_metadata['aggregation_statistic'], parameter=ds_metadata['parameter'], units=ds_metadata['units'], feature=ds_metadata['feature'], method=ds_metadata['method'], owner=ds_metadata['owner'])
 
     run_date_key = make_run_date_key(run_date)
-    ds1.attrs = {'featureType': 'timeSeries', 'title': title_str, 'institution': ds_mapping['owner'], 'license': ds_mapping['license'], 'source': ds_mapping['method'], 'history': run_date_key + ': Generated', 'version': 3}
+    data1.attrs = {'featureType': 'timeSeries', 'title': title_str, 'institution': ds_metadata['owner'], 'license': ds_metadata['license'], 'source': ds_metadata['method'], 'history': run_date_key + ': Generated', 'version': 3}
 
     ## Test conversion to netcdf
-    p_ts1 = ds1.to_netcdf()
+    p_ts1 = data1.to_netcdf()
 
     ## Compress if requested
     if compression:
@@ -336,7 +268,168 @@ def data_to_xarray(results_data, station_data, param_name, results_attrs, result
 
         return c_obj
     else:
-        return ds1
+        return data1
+
+
+# def data_to_xarray(results_data, station_data, param_name, attrs, encoding, station_attrs=None, station_encoding=None, run_date=None, ancillary_variables=None, compression=False, compress_level=1):
+#     """
+#     Converts DataFrames of time series data, station data, and other attributes to an Xarray Dataset. Optionally has Zstandard compression.
+
+#     Parameters
+#     ----------
+#     results_data : DataFrame
+#         DataFrame of the core parameter and associated ancillary variable. If the spatial distribution is sparse then the data should be indexed by geometry, time, and height. If the spatial distribution is grid then the data should be indexed by lon, lat, time, and height.
+#     station_data : dict
+#         Dictionary of the station data. Should include a station_id which should be a hashed string from blake2b (digest_size=12) of the geojson geometry. The other necessary field is geometry, which is the geometry of the results object is not grouped or a boundary extent as a polygon if the results object is grouped. Data owner specific other fields can include "ref" for the reference id and "name" for the station name.
+#     param_name : str
+#         The core parameter name of the column in the results_data DataFrame.
+#     results_attrs : dict
+#         A dictionary of the xarray/netcdf attributes of the results_data. Where the keys are the columns and the values are the attributes. Only necessary if additional ancilliary valiables are added to the results_data.
+#     results_encoding : dict
+#         A dictionary of the xarray/netcdf encodings for the results_data.
+#     station_attrs : dict or None
+#         Similer to results_attr, but can be omitted if no extra fields are included in station_data.
+#     station_encoding : dict or None
+#         Similer to results_encoding, but can be omitted if no extra fields are included in station_data.
+
+#     Returns
+#     -------
+#     Xarray Dataset or bytes object
+#     """
+#     ## dataset metadata
+#     ds_mapping = results_attrs[param_name]
+#     param_name = ds_mapping['parameter']
+#     sd = ds_mapping['spatial_distribution']
+#     grouping = ds_mapping['grouping']
+#     geo_type = ds_mapping['geometry_type']
+
+#     ## Integrity Checks
+
+#     # Time series data
+#     ts_data1 = results_data_integrety_checks(results_data, param_name, results_attrs, results_encoding, sd, ancillary_variables)
+
+#     # Station data
+#     stn_data = station_data_integrety_checks(station_data, attrs=station_attrs, encoding=station_encoding)
+
+#     # Metadata
+#     metadata_integrety_checks(ts_data1, stn_data, ds_mapping)
+
+#     ## geometry from station data
+#     stn_geometry = stn_data['geometry']
+#     stn_data.pop('geometry')
+#     if grouping == 'blocks':
+#         stn_data['extent'] = shape(stn_geometry).wkb_hex
+
+#     ## Station properties
+#     if 'properties' in stn_data:
+#         props = stn_data['properties'].copy()
+#         for k, v in props.items():
+#             stn_data[k] = v
+#         stn_data.pop('properties')
+
+#     ## Assign Attributes
+#     attrs1 = {'station_id': {'cf_role': "timeseries_id"}, 'lat': {'standard_name': "latitude", 'units': "degrees_north"}, 'lon': {'standard_name': "longitude", 'units': "degrees_east"}, 'altitude': {'standard_name': 'surface_altitude', 'long_name': 'height above the geoid to the lower boundary of the atmosphere', 'units': 'm'}, 'geometry': {'long_name': 'The hexadecimal encoding of the Well-Known Binary (WKB) geometry', 'crs_EPSG': 4326}, 'extent': {'long_name': 'The hexadecimal encoding of the Well-Known Binary (WKB) station extent', 'crs_EPSG': 4326}}
+
+#     if 'name' in stn_data.keys():
+#         attrs1.update({'name': {'long_name': 'station name'}})
+#     if 'ref' in stn_data.keys():
+#         attrs1.update({'ref': {'long_name': 'station reference id given by the owner'}})
+
+#     if isinstance(station_attrs, dict):
+#         attrs1.update(copy.deepcopy(station_attrs))
+
+#     ts_cols = list(ts_data1.columns)
+
+#     attrs1.update(results_attrs)
+#     if 'cf_standard_name' in attrs1[param_name]:
+#         attrs1[param_name]['standard_name'] = attrs1[param_name].pop('cf_standard_name')
+#     attrs1.update({'height': {'standard_name': 'height', 'long_name': 'vertical distance above the surface', 'units': 'm', 'positive': 'up'}, 'time': {'standard_name': 'time', 'long_name': 'start_time'}})
+#     if 'modified_date' in ts_cols:
+#         attrs1.update({'modified_date': {'long_name': 'last modified date'}})
+
+#     if isinstance(ancillary_variables, list):
+#         if len(ancillary_variables) > 0:
+#             attrs1[param_name].update({'ancillary_variables': ' '.join(ancillary_variables)})
+
+#     ## Assign encodings
+#     encoding1 = {'lon': {'dtype': 'int32', '_FillValue': -999999, 'scale_factor': 0.0000001}, 'lat': {'dtype': 'int32', '_FillValue': -999999, 'scale_factor': 0.0000001}, 'altitude': {'dtype': 'int32', '_FillValue': -9999, 'scale_factor': 0.001}}
+
+#     if isinstance(station_encoding, dict):
+#         encoding1.update(copy.deepcopy(station_encoding))
+
+#     height = pd.to_numeric(ts_data1.reset_index()['height'], downcast='integer')
+
+#     if 'int' in height.dtype.name:
+#         height_enc = {'dtype': height.dtype.name}
+#     elif 'float' in height.dtype.name:
+#         height_enc = {'dtype': 'int32', '_FillValue': -9999, 'scale_factor': 0.001}
+#     else:
+#         raise TypeError('height should be either an int or a float')
+
+#     encoding1.update(results_encoding)
+#     encoding1.update({'time': {'_FillValue': -99999999, 'units': "days since 1970-01-01 00:00:00"}, 'height': height_enc})
+#     if 'modified_date' in ts_cols:
+#         encoding1.update({'modified_date': {'_FillValue': -99999999, 'units': "days since 1970-01-01 00:00:00"}})
+
+#     ## Create the Xarray Dataset
+#     data_index = list(ts_data1.index.names)
+#     ds1 = ts_data1.to_xarray()
+
+#     ## Assign the stn data to the main Dataset
+#     if 'extent' in stn_data:
+#         ds1 = ds1.assign_coords({'extent': [stn_data['extent']]})
+#         stn_data.pop('extent')
+
+#     if 'extent' in ds1:
+#         stn_index = ('extent')
+#     else:
+#         if 'geometry' in data_index:
+#             stn_index = ('geometry')
+#         else:
+#             stn_index = ('lon', 'lat')
+
+#     # ds1 = ds1.assign_coords({'station_id': (stn_index, [stn_data['station_id']])})
+#     # stn_data.pop('station_id')
+
+#     for k, v in stn_data.items():
+#         ds1 = ds1.assign({k: (stn_index, [v])})
+
+#     ## Assign the lat and lon if the data are sparse points
+#     if ('geometry' in data_index) and (stn_geometry['type'] == 'Point'):
+#         ds1 = ds1.assign({'lon': (stn_index, [stn_geometry['coordinates'][0]])})
+#         ds1 = ds1.assign({'lat': (stn_index, [stn_geometry['coordinates'][1]])})
+
+#     ## Add attributes and encodings
+#     for e, val in encoding1.items():
+#         if e in ds1:
+#             if ('dtype' in val) and (not 'scale_factor' in val):
+#                 if 'int' in val['dtype']:
+#                     ds1[e] = ds1[e].astype(val['dtype'])
+#             if 'scale_factor' in val:
+#                 precision = int(np.abs(np.log10(val['scale_factor'])))
+#                 ds1[e] = ds1[e].round(precision)
+#             ds1[e].encoding = val
+
+#     for a, val in attrs1.items():
+#         if a in ds1:
+#             ds1[a].attrs = val
+
+#     title_str = '{agg_stat} {parameter} in {units} of the {feature} by a {method} owned by {owner}'.format(agg_stat=ds_mapping['aggregation_statistic'], parameter=ds_mapping['parameter'], units=ds_mapping['units'], feature=ds_mapping['feature'], method=ds_mapping['method'], owner=ds_mapping['owner'])
+
+#     run_date_key = make_run_date_key(run_date)
+#     ds1.attrs = {'featureType': 'timeSeries', 'title': title_str, 'institution': ds_mapping['owner'], 'license': ds_mapping['license'], 'source': ds_mapping['method'], 'history': run_date_key + ': Generated', 'version': 3}
+
+#     ## Test conversion to netcdf
+#     p_ts1 = ds1.to_netcdf()
+
+#     ## Compress if requested
+#     if compression:
+#         cctx = zstd.ZstdCompressor(level=compress_level)
+#         c_obj = cctx.compress(p_ts1)
+
+#         return c_obj
+#     else:
+#         return ds1
 
 
 def compare_dfs(old_df, new_df, on, parameter, add_old=False):
@@ -542,7 +635,7 @@ def process_datasets(datasets):
     return dataset_list
 
 
-def create_geometry_df(df, extent=False, altitude=True, precision=7):
+def create_geometry_df(df, extent=False, altitude=True, to_wkb_hex=False, precision=7):
     """
 
     """
@@ -555,6 +648,8 @@ def create_geometry_df(df, extent=False, altitude=True, precision=7):
             geometry = geojson.Polygon([[(min_lon, min_lat), (min_lon, max_lat), (max_lon, max_lat), (max_lon, min_lat), (min_lon, min_lat)]], True, precision=precision)
         else:
             raise ValueError('Extent must have lat and lon in the df.')
+        if to_wkb_hex:
+                geometry = shape(geometry).wkb_hex
     else:
         if 'geometry' in df:
             geometry = df['geometry']
@@ -570,6 +665,9 @@ def create_geometry_df(df, extent=False, altitude=True, precision=7):
         else:
             raise ValueError('Either a dict of geometry or a combo of lat and lon must be in the dataframe.')
 
+        if to_wkb_hex:
+            geometry = geometry.apply(lambda x: shape(x).wkb_hex)
+
     return geometry
 
 
@@ -630,36 +728,36 @@ def get_new_stats(data):
 #     return stns2
 
 
-def process_station_base(coords=None, station_id=None, geometry=None, ref=None, name=None, osm_id=None, altitude=None, properties=None, return_dict=True):
-    """
+# def process_station_base(coords=None, station_id=None, geometry=None, ref=None, name=None, osm_id=None, altitude=None, properties=None, return_dict=True):
+#     """
 
-    """
-    if isinstance(coords, list):
-        ## Create geometry and station_id
-        geometry = geojson.Point(coords, True, precision=5)
-        station_id = assign_station_id(geometry)
-    elif (not isinstance(station_id, str)) & (not isinstance(geometry, dict)):
-        raise ValueError('coords or station_id must be assigned')
+#     """
+#     if isinstance(coords, list):
+#         ## Create geometry and station_id
+#         geometry = geojson.Point(coords, True, precision=5)
+#         station_id = assign_station_id(geometry)
+#     elif (not isinstance(station_id, str)) & (not isinstance(geometry, dict)):
+#         raise ValueError('coords or station_id must be assigned')
 
-    ## Put into data model
-    stn_m = StationBase(station_id=station_id, geometry=geometry, ref=ref, name=name, osm_id=osm_id, altitude=altitude, properties=properties)
+#     ## Put into data model
+#     stn_m = StationBase(station_id=station_id, geometry=geometry, ref=ref, name=name, osm_id=osm_id, altitude=altitude, properties=properties)
 
-    if return_dict:
-        stn_m = orjson.loads(stn_m.json(exclude_none=True))
+#     if return_dict:
+#         stn_m = orjson.loads(stn_m.json(exclude_none=True))
 
-    return stn_m
+#     return stn_m
 
 
-def process_stations_base(stns_list):
-    """
+# def process_stations_base(stns_list):
+#     """
 
-    """
-    stns_dict = {}
-    for s in stns_list:
-        stn_m = process_station_base(**s)
-        stns_dict[stn_m['station_id']] = stn_m
+#     """
+#     stns_dict = {}
+#     for s in stns_list:
+#         stn_m = process_station_base(**s)
+#         stns_dict[stn_m['station_id']] = stn_m
 
-    return stns_dict
+#     return stns_dict
 
 
 def get_station_data_from_xr(data):
@@ -680,19 +778,22 @@ def get_station_data_from_xr(data):
     stn_fields = list(StationBase.schema()['properties'].keys())
 
     ## Geometry
-    if 'geometry' in dims1:
-        if len(data['geometry']) == 1:
-            geo1 = mapping(wkb.loads(data['geometry'].values[0], True))
-        else:
-            geo1 = mapping(wkb.loads(data['extent'].values[0], True))
+    if 'extent' in dims1:
+        geo1 = mapping(wkb.loads(data['extent'].values[0], True))
+    elif 'geometry' in dims1:
+        geo1 = mapping(wkb.loads(data['geometry'].values[0], True))
+    else:
+        lon = data['lon'].values[0]
+        lat = data['lat'].values[0]
+        geo1 = geojson.Point([lon, lat], True, 7)
 
-        stn_fields.remove('geometry')
+    stn_fields.remove('geometry')
 
     lat_lon = ['lon', 'lat']
 
     stn_vars = [v for v in vars1 if (not v in dims1) and (not v in data_vars) and (not v in lat_lon)]
     stn_data1 = {k: v['data'][0] for k, v in data[stn_vars].to_dict()['data_vars'].items() if k in stn_fields}
-    stn_data1.update({'station_id': data['station_id'].values[0], 'geometry': geo1})
+    stn_data1.update({'geometry': geo1})
     if 'altitude' in stn_data1:
         stn_data1['altitude'] = round(stn_data1['altitude'], 3)
     # if not 'virtual_station' in stn_data1:
@@ -732,23 +833,41 @@ def process_station_summ(dataset_id, station_id, data, object_infos, mod_date=No
     return stn_dict
 
 
-def prepare_results(data_dict, dataset_list, station_data, results_data, run_date_key, mod_date=None, sum_closed='right', other_closed='left', discrete=True, station_attrs=None, station_encoding=None):
+def prepare_results(data_dict, dataset_list, results_data, run_date_key, sum_closed='right', other_closed='left', discrete=True, other_attrs=None, other_encoding=None):
     """
 
     """
-    if isinstance(mod_date, (str, pd.Timestamp)):
-        mod_date = pd.Timestamp(mod_date)
-        ancillary_variables = ['modified_date']
-    else:
-        ancillary_variables = []
+    # if :
+    #     mod_date = pd.Timestamp(mod_date)
+    #     ancillary_variables = ['modified_date']
+    # else:
+    #     ancillary_variables = []
 
     tz_str = 'Etc/GMT{0:+}'
 
-    ## Determine index
-    if not isinstance(results_data.index, pd.MultiIndex):
-        raise TypeError('The data_df index must be a MultiIndex.')
+    parameter = dataset_list[0]['parameter']
 
-    data_index = list(results_data.index.names)
+    if parameter not in results_data:
+        raise ValueError('The parameter ' + str(parameter) + ' is not in the results_data.')
+
+    ## Determine index
+    data_index = tuple(results_data.dims)
+    vars2 = list(results_data.variables)
+    vars3 = [v for v in vars2 if v not in data_index]
+
+    vars_dict = {}
+    for v in vars3:
+        index1 = results_data[v].dims
+        vars_dict[v] = index1
+
+    ancillary_variables = [v for v, i in vars_dict.items() if (i == data_index) and (v != parameter)]
+
+    main_vars = [parameter] + ancillary_variables
+
+    # if not isinstance(results_data.index, pd.MultiIndex):
+    #     raise TypeError('The data_df index must be a MultiIndex.')
+
+    # data_index = list(results_data.index.names)
 
     if not 'time' in data_index:
         raise ValueError('time must be in the data_df index.')
@@ -770,67 +889,70 @@ def prepare_results(data_dict, dataset_list, station_data, results_data, run_dat
         attrs1 = copy.deepcopy(attrs)
         attrs1.update({ds_mapping['parameter']: ds_mapping})
 
+        if isinstance(other_attrs, dict):
+            attrs1.update(other_attrs)
+
         encoding1 = copy.deepcopy(encoding)
 
+        if isinstance(other_encoding, dict):
+            encoding1.update(other_encoding)
+
         ## Pre-Process data
         qual_col = 'quality_code'
         freq_code = ds_mapping['frequency_interval']
         utc_offset = ds_mapping['utc_offset']
-        parameter = ds_mapping['parameter']
 
-        main_cols = data_index.copy()
+        main_cols = list(data_index)
         main_cols.extend([parameter])
 
-        ts_data1 = results_data.reset_index().copy()
+        ts_data1 = results_data.copy()
 
         # Convert times to local TZ if necessary
         if (not freq_code in ['T', 'H', '1H']) and (not utc_offset == '0H'):
             t1 = int(utc_offset.split('H')[0])
             tz1 = tz_str.format(-t1)
-            ts_data1['time'] = ts_data1['time'].dt.tz_localize('UTC').dt.tz_convert(tz1).dt.tz_localize(None)
+            ts_data1['time'] = ts_data1['time'].to_index().tz_localize('UTC').tz_convert(tz1).tz_localize(None)
 
         ## Aggregate data if necessary
         # Parameter
         if freq_code == 'T':
-            grp1 = ts_data1.groupby(data_index)
-            data1 = grp1[parameter].mean()
-
+            data2 = ts_data1
         else:
             agg_fun = agg_stat_mapping[ds_mapping['aggregation_statistic']]
 
+            ts_data2 = ts_data1[main_vars].to_dataframe()
+
             if agg_fun == 'sum':
-                data1 = grp_ts_agg(ts_data1[main_cols], other_index, 'time', freq_code, agg_fun, closed=sum_closed)
+                data0 = grp_ts_agg(ts_data2[parameter].reset_index(), other_index, 'time', freq_code, agg_fun, closed=sum_closed)
             else:
-                data1 = grp_ts_agg(ts_data1[main_cols], other_index, 'time', freq_code, agg_fun, discrete, closed=other_closed)
+                data0 = grp_ts_agg(ts_data2[parameter].reset_index(), other_index, 'time', freq_code, agg_fun, discrete, closed=other_closed)
 
-        # Quality code
-        if qual_col in ts_data1.columns:
-            ts_data1[qual_col] = pd.to_numeric(ts_data1[qual_col], errors='coerce', downcast='integer')
-            if freq_code == 'T':
-                qual1 = grp1[qual_col].min()
-            else:
-                qual_cols = data_index.copy()
-                qual_cols.extend([qual_col])
-                qual1 = grp_ts_agg(ts_data1[qual_cols], other_index, 'time', freq_code, 'min')
-            df3 = pd.concat([data1, qual1], axis=1).reset_index().dropna()
+            # Ancillary variables
+            av_list = [data0]
+            for av in ancillary_variables:
+                if qual_col == av:
+                    ts_data2[qual_col] = pd.to_numeric(ts_data2[qual_col], errors='coerce', downcast='integer')
+                    qual_cols = data_index.copy()
+                    qual_cols.extend([qual_col])
+                    qual1 = grp_ts_agg(ts_data2[qual_cols].reset_index(), other_index, 'time', freq_code, 'min')
+                    av_list.append(qual1)
+                else:
+                    av1 = grp_ts_agg(ts_data2[av].reset_index(), other_index, 'time', freq_code, 'max')
+                    av_list.append(av1)
 
-            ancillary_variables.extend([qual_col])
-        else:
-            df3 = data1.reset_index().copy()
-
-        if 'modified_date' in ancillary_variables:
-            df3['modified_date'] = mod_date
-
-        ## Convert to xarray
-        df4 = df3.copy()
+            # Put the data together
+            data1 = pd.concat(av_list, axis=1)
+            data2 = data1.to_xarray()
+            for v in vars3:
+                if v not in data2:
+                    data2[v] = ts_data1[v]
 
         # Convert time back to UTC if necessary
         if (not freq_code in ['T', 'H', '1H']) and (not utc_offset == '0H'):
-            df4['time'] = df4['time'].dt.tz_localize(tz1).dt.tz_convert('utc').dt.tz_localize(None)
+            data2['time'] = data2['time'].to_index().tz_localize(tz1).tz_convert('utc').tz_localize(None)
 
-        df4.set_index(data_index, inplace=True)
-
-        new1 = data_to_xarray(df4, station_data, parameter, attrs1, encoding1, station_attrs=station_attrs, station_encoding=station_encoding, run_date=run_date_key, ancillary_variables=ancillary_variables, compression='zstd')
+        ## Package up the netcdf object
+        new1 = package_xarray(data2, parameter, attrs1, encoding1, run_date=run_date_key, compression='zstd')
 
         ## Update the data_dict
         ds_id = ds_mapping['dataset_id']
