@@ -324,7 +324,6 @@ def compare_datasets_from_s3(conn_config, bucket, new_data, add_old=False, last_
     base_key_pattern = key_patterns['results']
 
     ## Get list of keys
-
     if isinstance(last_run_date_key, str):
         key_dict.update({'run_date': last_run_date_key})
         last_key = base_key_pattern.format(**key_dict)
@@ -424,9 +423,9 @@ def update_results_s3(processing_code, data_dict, run_date_dict, remote, threads
 
             vars1 = list(new1.variables)
             parameter = [v for v in vars1 if 'dataset_id' in new1[v].attrs][0]
-            attrs = new1[parameter].attrs.copy()
+            result_attrs = new1[parameter].attrs.copy()
 
-            ds_id = attrs['dataset_id']
+            ds_id = result_attrs['dataset_id']
             mod_date_key = new1.attrs['history'].split(':')[0]
 
             if no_comparison:
